// ==========================================================
//  
// CBP_3352_Salamander_1.c
// 03-02-21 Controller for the Salamander CPG
//
// ==========================================================

#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/gps.h>
#include <webots/distance_sensor.h>
 
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

// (1) ODE Solver includes
#include "rkf45.h"
#include "CBP_rhs_Salamander_1.h"
#include "CBP_Helper_H.h"

#define TIME_STEP 16

// (2) Motors include 6 segments and 4 legs (not used here)
#define NUM_MOTORS 10

// (3) Number of ODEs in the CPG
#define NEQN 17 
#define DEBUG false

  // (4) segments and legs in a single array for coding efficiency
  WbDeviceTag motor[NUM_MOTORS];
  double seg[NUM_MOTORS] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};

 // (5) distance sensors (left and right eyes) and gps
  WbDeviceTag eyeL_tag, eyeR_tag, gps; 

  // (6) get the motors device tags from the PROTO
  char *MOTOR_NAMES[NUM_MOTORS] = {"motor_1", "motor_2","motor_3","motor_4","motor_5",
                                   "motor_6", "motor_leg_1", "motor_leg_2", "motor_leg_3","motor_leg_4"};

// (7) motors angle range 
static double min_motor_angle[NUM_MOTORS];
static double max_motor_angle[NUM_MOTORS];

// (8) Function to limit the motor angles
double clamp(double value, double min, double max) {
  if (min > max) {
    assert(0);
    return value;
  } else if (min == 0 && max == 0) {
    return value;
  }

  return value < min ? min : value > max ? max : value;
}

void init() {
  // (9) Get min and max allowed angles for all segments and legs
  for (int i = 0; i < NUM_MOTORS; i++) {
    motor[i] = wb_robot_get_device(MOTOR_NAMES[i]);
    min_motor_angle[i] = wb_motor_get_min_position(motor[i]);
    max_motor_angle[i] = wb_motor_get_max_position(motor[i]);
  }

  // (10) get and enable left and right distance sensors 
  eyeL_tag = wb_robot_get_device("ds_left");
  wb_distance_sensor_enable(eyeL_tag, TIME_STEP);
  eyeR_tag = wb_robot_get_device("ds_right");
  wb_distance_sensor_enable(eyeR_tag, TIME_STEP);
  
  // (11) get and enable gps device 
  gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, TIME_STEP);

}


int main(int argc, char **argv) {

  // (12) Variable related to the ODE solver
  double abserr;
  int flag;
  double relerr;
  double t;
  double t_out;
  double y[NEQN];
  double dydt[NEQN];
  
  double pi = M_PI;
  
  // (13) params for this model
  double alpha;
  double mu;
  double omega;
  double dPhase;
  double tau,k;
  //double scaling;
  double ampl;
   
  // (14) GPS stuff to calculate the salamander's speed
  double x = 0;
  double yy = 0;
  double xPrev, yyPrev, speed;

  // (15) Logging files
  const char *name;  
  char fname[10];  
  FILE* logFile;
  FILE* exptFile;
    
  int i;

  wb_robot_init(); 
  
  init();
  
  // (16) Get the phase lag, the frequency and the amplitude from contollerArgs
  // in the scene tree
  dPhase = atof(argv[1]); // in degrees
  omega  = atof(argv[2]); // in Herts
  ampl   = atof(argv[3]); 
  
  // (17) Fix these parameters
  alpha = 1;
  mu = 1;
  
  // (18) Calculate tau and k to get correct phase lag spread over 3 neurons
  dPhase = (M_PI/180.0)*dPhase;
  omega = omega*2*pi;
  tau = tan(dPhase/3.0)/omega;
  k = sqrt(tau*tau*omega*omega + 1);

  printf("dPhase = %f, omega = %f, tau = %f, k = %f\n",dPhase,omega,tau,k);
 
  // (19) Get robot name and open up some logging files
  name = wb_robot_get_name();
  sprintf(fname,"Robot_%s.m",name);
  logFile = fopen(fname,"w");
  
  writeOctaveHead(logFile);
  
  sprintf(fname,"%s.txt",name);
  printf("%s\n",fname);
  
  exptFile = fopen(fname,"a");
  
  // (20) Log the cvurrent parameters to the exptFile
  fprintf(exptFile,"----------------------------------------\n"); 
  fprintf(exptFile,"omega  = %f\n",omega);  
  fprintf(exptFile,"dPhase  = %f (%2f)\n",dPhase,dPhase*180/M_PI);
  fprintf(exptFile,"tau    = %f\n",tau);
  fprintf(exptFile,"k      = %f\n",k);   
  fprintf(exptFile,"\n"); 
  
 // (21) Calculate the epsilon - numerical precision for this computer. Use in
 // the solver
  abserr = 100*sqrt ( r8_epsilon ( ) );
  relerr = 100*sqrt ( r8_epsilon ( ) );
  printf("Root epsilon: %e\n",abserr);
  
  // (22) Initialize the solver
  flag = 1;
  // Set some initial conditions
  for(i=0; i < 17; i++)
    y[i] = 0.5;//((double)rand()/(double)RAND_MAX);
  
  // (23) Send the parameters to the solver
  setParams(alpha, mu, omega,tau,k);
	
  // (24) Initialize the ODE Solver                 
  t = 0.0;
  t_out = TIME_STEP/1000.0; 
  rhs_Salamander_1( t, y,  dydt );
    
  // Main while loop      
  while (wb_robot_step(TIME_STEP) != -1) {
                    
    // (25) Call to the ODE solver rhs_Salamander_1.c  
    flag = r8_rkf45 ( rhs_Salamander_1, NEQN, y, dydt, &t, t_out, &relerr, abserr, flag );
    // (26) Update the current time t_out
    t_out = t + TIME_STEP/1000.0;
    
    // (27) Print out any warning flags we get from the solver
    if(flag == 4) {
      printf("WARNING: flag 4 at t = %f\n",t);
      flag = 2;
    }
    if(flag == 7) {
      printf("WARNING: flag 7 at t = %f\n",t);
      flag = 2; 
    }   
    if(flag == 6) {
      abserr *=1.1;
      relerr *=1.1;
      flag = 2;
      printf("WARNING: flag 6: abserr = %f\n",abserr);
    }
   
    // (28) GPS for speed calc    
    xPrev = x;
    yyPrev = yy;
    const double *gpsVals = wb_gps_get_values(gps);  
    x = gpsVals[0];
    yy = gpsVals[2];    
    speed = sqrt(pow(x - xPrev,2) + pow(yy - yyPrev,2))/(TIME_STEP/1000.0);
    printf("speed = %f\n",speed); 

    // (29) Mapping from the neuron array y[...] to the salamander segment motors    
     seg[0] = ampl*y[0];
     seg[1] = ampl*y[4];     
     seg[2] = ampl*y[7];     
     seg[3] = ampl*y[10];
     seg[4] = ampl*y[13];
     seg[5] = ampl*y[16];
                  
    // (30) Drive the Salamander motors
    for (int i = 0; i < 6; i++) {
      seg[i] = clamp(seg[i], min_motor_angle[i], max_motor_angle[i]);
      wb_motor_set_position(motor[i], seg[i]);
    }	
    
    writeData6ToOctave(logFile,t_out,seg[0],seg[1],seg[2],seg[3],seg[4],seg[5]);	
	 
 }
  
  // (31) Flush all pending printf's to the files and close them
  writeOctaveTail(logFile);
  fflush(logFile);
  fclose(logFile);
  fflush(exptFile);
  fclose(exptFile);

  wb_robot_cleanup();

  return 0;
}
